;%include "asm.inc"
segment .text

%macro assume 2
	%define %2(x) [%1 + %2. %+ x]
%endm

%define sizeof(x) x %+ _size

struc CONTEXT
.ctxflg resd 1+6+28+4
;Dr0    resd 1
;Dr1    resd 1
;Dr2    resd 1
;Dr3    resd 1
;Dr6    resd 1
;Dr7    resd 1
;fltsav resd 28

;gs     resd 1+3
;fs     resd 1
;es     resd 1
;ds     resd 1

.edi    resd 1
.esi    resd 1
.ebx    resd 1
.edx    resd 1
.ecx    resd 1
.eax    resd 1

.ebp    resd 1
.eip    resd 1+2
;cs     resd 1
;eFlags resd 1
.esp    resd 1+2
;ss     resd 1
;ExtendedRegisters resd 1
endstruc

struc EXCEPTION_RECORD
.ExceptionCode			resd 1
.ExceptionFlags			resd 1
.pExceptionRecord		resd 1
.ExceptionAddress		resd 1
.NumberParameters		resd 1
.ExceptionInformation	resd 1
endstruc

%define EXCEPTION_USER_THROW -2

global __try
global __catch
global __leave
global __throw

struc EXCEPTION_REGISTRATION
.next     resd 1
.handler  resd 1
; user defined
.resume   resd 1 ; if __try() {} <- here, outside IF
.ebp      resd 1 ; saved ebp
.esp      resd 1 ; saved esp
.code     resd 1 ; ExceptionCode
.addr     resd 1 ; ExceptionAddress
.userdata resd 1 ; used by __throw
endstruc

;-----------------------------------------------------------
align 4
__try:
	pop  eax           ; address of 'test eax, eax'
	lea  ecx,[eax+8]   ; address after 'if __try'
	add  eax,2         ; address of jz near
	add  eax,[eax+2]   ; add the offset from jz near
	add  eax,6         ; skip the jz near
	xor  edx,edx
	push edx                ; EXCEPTION_REGISTRATION.addr
	push edx                ; EXCEPTION_REGISTRATION.code
	push esp                ; EXCEPTION_REGISTRATION.esp   ; invalid!
	push ebp                ; EXCEPTION_REGISTRATION.ebp
	push eax                ; EXCEPTION_REGISTRATION.resume
	push _exception_handler ; EXCEPTION_REGISTRATION.handler
	push dword[fs:0]        ; EXCEPTION_REGISTRATION.next
	mov  [fs:0],esp         ; add new node to current thread SEH

	mov  [esp+EXCEPTION_REGISTRATION.esp],esp ; fix the invalid esp
	jmp  ecx                ; jump (return) inside IF


;-----------------------------------------------------------

align 4
__leave:
	pop    ecx          ; pop return address
	xchg   edx,esi      ; save esi in edx
	mov    esi,[fs:0]
	assume esi,EXCEPTION_REGISTRATION

	and    esi,esi
	jz    .q            ; cancel if [fs:0] is NULL
	inc    esi
	jz    .q            ; or -1
	dec    esi
	cmp    dword EXCEPTION_REGISTRATION(handler),_exception_handler
	jne   .q
	pop    dword[fs:0] ; restore previous handler
	add    esp,sizeof(EXCEPTION_REGISTRATION)
.q:
	xchg   edx,esi      ; restore esi
	jmp    ecx          ; and return

;-----------------------------------------------------------

align 4
__catch:                ; code*, addr*, string*
	enter  0,0
	push   edi
	push   esi
	mov    esi,[fs:0]
	assume esi,EXCEPTION_REGISTRATION

	xor    eax,eax      ; default return value is FALSE

	and    esi,esi
	jz    .q            ; return FALSE if [fs:0] is NULL
	inc    esi
	jz    .q            ; or -1
	dec    esi
	cmp    dword EXCEPTION_REGISTRATION(handler),_exception_handler
	jne   .q            ; or the handler is not equal to _exception_handler()

	mov    ecx,EXCEPTION_REGISTRATION(code)
	mov    eax,EXCEPTION_REGISTRATION(addr)
	and    ecx,ecx
	jz    .q                ; here was no exception
	mov    edi,[ebp+8]
	and    edi,edi
	jz    .sk1
	mov    [edi],ecx
	.sk1:
	mov    edi,[ebp+12]
	and    edi,edi
	jz    .sk2
	mov    [edi],eax
	.sk2:

	; search for exception string (code in edx)
	call   _get_exception_string

	mov    edi,[ebp+16]
	and    edi,edi
	jz    .sk3
	mov    [edi],eax
	.sk3:
	mov    al,1
.q:
	pop    esi
	pop    edi
	leave
;	ret    12
	pop    edx          ; return address
	add    esp,12
	; we are now in caller stack frame. Imitating __leave() call from the caller...
	push   edx
	jmp    __leave      ; don't worry about eax

;-----------------------------------------------------------
align 4
_exception_handler: ; EXCEPTION_RECORD*, EXCEPTION_REGISTRATION*, CONTEXT*, void *DispatcherContext)
	enter  0,0
	push   esi
	push   edi
	mov    edi,[ebp+12] ; EXCEPTION_REGISTRATION
	mov    esi,[ebp+16] ; CONTEXT *

	assume edi,EXCEPTION_REGISTRATION
	assume esi,CONTEXT

	mov    eax,EXCEPTION_REGISTRATION(esp)
	mov    ecx,EXCEPTION_REGISTRATION(ebp)
	mov    edx,EXCEPTION_REGISTRATION(resume)

	mov    CONTEXT(esp),eax
	mov    CONTEXT(ebp),ecx
	mov    CONTEXT(eip),edx


	mov    esi,[ebp+8] ; EXCEPTION_RECORD
	assume esi,EXCEPTION_RECORD

	; exception generated in __throw.allowed ?
	cmp    dword EXCEPTION_RECORD(ExceptionAddress), __throw.allowed
	je    .skip ; yes! do not change exception code or address

	mov    eax,EXCEPTION_RECORD(ExceptionCode)
	mov    ecx,EXCEPTION_RECORD(ExceptionAddress)
	mov    EXCEPTION_REGISTRATION(code),eax
	mov    EXCEPTION_REGISTRATION(addr),ecx
.skip:

	pop    edi
	pop    esi
	leave
	xor    eax,eax
	ret    16
;

;-----------------------------------------------------------
align 4
_get_exception_string: ; ecx = code
	cmp   ecx, EXCEPTION_USER_THROW
	jne  .f
	mov   eax,$g_user_message
	ret
.f:
	push  esi
	mov   esi,_e_codes
.f1:
	lodsd
	cmp    eax,0
	je    .fq
	cmp    eax,ecx
	je    .fq
.f2:
	; skip string
	lodsb
	and    al,al
	jnz    .f2
	; align
	mov    eax,4
	sub    eax,esi
	and    eax,3
	add    esi,eax
	jmp   .f1
.fq:
	mov    eax,esi
	pop    esi
	ret

;-----------------------------------------------------------
align 4
__throw:   ; string *
	push   esi
	push   edi
	mov    esi,[fs:0]

	assume esi,EXCEPTION_REGISTRATION

	and    esi,esi
	jz    .q            ; ignore if [fs:0] is NULL
	inc    esi
	jz    .q            ; or -1
	dec    esi
	cmp    dword EXCEPTION_REGISTRATION(handler),_exception_handler
	jne   .q            ; or the handler is not equal to _exception_handler()

	mov    dword EXCEPTION_REGISTRATION(code),EXCEPTION_USER_THROW

	mov    eax,[esp+8]  ; save return address in SEH frame
	mov    EXCEPTION_REGISTRATION(addr),eax
	; copy the string
	mov    esi,[esp+12]   ; string *
	and    esi,esi
	jz    .q              ; ignore if NULL
	mov    edi,$g_user_message
	mov    cl,255
.c:
	lodsb
	mov    [edi],al
	add    edi,1
	and    al,al
	jz    .exec
	sub    cl,1
	jnz   .c
	mov    byte[edi],0
.exec:
	pop    eax ; edi
	pop    eax ; esi
	pop    eax ; return address
	pop    eax ; string *
	xor    eax,eax
.allowed
	mov    [eax],eax  ; generate exception
.q:
	pop    edi
	pop    esi
	ret    4

;-----------------------------------------------------------

%define DBG_EXCEPTION_HANDLED            (0x00010001)
%define DBG_CONTINUE                     (0x00010002)
%define DBG_TERMINATE_THREAD             (0x40010003)
%define DBG_TERMINATE_PROCESS            (0x40010004)
%define DBG_COMMAND_EXCEPTION            (0x40010009)
%define DBG_EXCEPTION_NOT_HANDLED        (0x80010001)
%define STATUS_SEGMENT_NOTIFICATION      (0x40000005)
%define STATUS_REG_NAT_CONSUMPTION       (0xC00002C9)
%define STATUS_SXS_EARLY_DEACTIVATION    (0xC015000F)
%define STATUS_SXS_INVALID_DEACTIVATION  (0xC0150010)

%macro EXCEPTION 3
	%push e
	dd %2
	db %3,0
	align 4
%$
	%pop
%endm

;-----------------------------------------------------------
segment .bss
$g_user_message: resb 256

;-----------------------------------------------------------
segment .data
_e_codes:
EXCEPTION DBG_CONTROL_BREAK               , (0x40010008),"Control-Break"
EXCEPTION DBG_CONTROL_C                   , (0x40010005),"Control-C"
EXCEPTION STATUS_GUARD_PAGE_VIOLATION     , (0x80000001), "Guard Page Violation"
EXCEPTION STATUS_DATATYPE_MISALIGNMENT    , (0x80000002), "Datatype Misalignment"
EXCEPTION STATUS_BREAKPOINT               , (0x80000003), "Breakpoint"
EXCEPTION STATUS_SINGLE_STEP              , (0x80000004), "Single Step"
EXCEPTION STATUS_ACCESS_VIOLATION         , (0xC0000005), "Access Violation"
EXCEPTION STATUS_IN_PAGE_ERROR            , (0xC0000006), "In Page Error"
EXCEPTION STATUS_INVALID_HANDLE           , (0xC0000008), "Invalid Handle"
EXCEPTION STATUS_NO_MEMORY                , (0xC0000017), "No Memory"
EXCEPTION STATUS_ILLEGAL_INSTRUCTION      , (0xC000001D), "Illegal Instruction"
EXCEPTION STATUS_NONCONTINUABLE_EXCEPTION , (0xC0000025), "Noncontinuable Exception"
EXCEPTION STATUS_INVALID_DISPOSITION      , (0xC0000026), "Invalid Disposition"
EXCEPTION STATUS_ARRAY_BOUNDS_EXCEEDED    , (0xC000008C), "Array Bounds Exceeded"
EXCEPTION STATUS_FLOAT_DENORMAL_OPERAND   , (0xC000008D), "Float Denormal Operand"
EXCEPTION STATUS_FLOAT_DIVIDE_BY_ZERO     , (0xC000008E), "Float Divide by Zero"
EXCEPTION STATUS_FLOAT_INEXACT_RESULT     , (0xC000008F), "Float Inexact Result"
EXCEPTION STATUS_FLOAT_INVALID_OPERATION  , (0xC0000090), "Float Invalid Operation"
EXCEPTION STATUS_FLOAT_OVERFLOW           , (0xC0000091), "Float Overflow"
EXCEPTION STATUS_FLOAT_STACK_CHECK        , (0xC0000092), "Float Stack Check"
EXCEPTION STATUS_FLOAT_UNDERFLOW          , (0xC0000093), "Float Underflow"
EXCEPTION STATUS_INTEGER_DIVIDE_BY_ZERO   , (0xC0000094), "Integer Divide by Zero"
EXCEPTION STATUS_INTEGER_OVERFLOW         , (0xC0000095), "Integer Overflow"
EXCEPTION STATUS_PRIVILEGED_INSTRUCTION   , (0xC0000096), "Privileged Instruction"
EXCEPTION STATUS_STACK_OVERFLOW           , (0xC00000FD), "Stack Overflow"
EXCEPTION STATUS_CONTROL_C_EXIT           , (0xC000013A), "Control-C Exit"
EXCEPTION STATUS_FLOAT_MULTIPLE_FAULTS    , (0xC00002B4), "Float Multiple Faults"
EXCEPTION STATUS_FLOAT_MULTIPLE_TRAPS     , (0xC00002B5), "Float Multiple Traps"
EXCEPTION 0,0,"Unknown Exception"

